import json, os, pickle as pkl
from .demand import DemandProfile
from .utils import *

def _save_objects(self, filename: str | None = None, overwrite: bool = True, json_indent: int = 4) -> None:
    """
    Save parameters of all TUD-SUMO objects created for the simulation (tracked edges/junctions, phases, controllers, events, demand, routes).

    Args:
        `filename` (str, optional): Output filename ('_.json_' or '_.pkl_')
        `overwrite` (bool): Denotes whether to allow overwriting previous outputs
        `json_indent` (int, optional): Indent used when saving JSON files
    """

    object_params = {}

    if len(self.tracked_edges) > 0:
        e_ids = list(self.tracked_edges.keys())
        object_params["edges"] = e_ids if len(e_ids) > 1 else e_ids[0]
    
    if len(self.tracked_junctions) > 0:
        object_params["junctions"] = {junc_id: junc._init_params for junc_id, junc in self.tracked_junctions.items()}

    if self._junc_phases != None:
        junc_phases = {}
        for junc_id, phase_dict in self._junc_phases.items():
            if junc_id in self.tracked_junctions and self.tracked_junctions[junc_id]._is_meter: continue
            junc_phases[junc_id] = {key: phase_dict[key] for key in ["phases", "times"]}
        if len(junc_phases) > 0:
            object_params["phases"] = junc_phases

    if len(self.controllers) > 0:
        object_params["controllers"] = {c_id: c._init_params for c_id, c in self.controllers.items()}

    if self._scheduler != None:
        events = self._scheduler.get_events()
        if len(events) > 0: object_params["events"] = {e.id: e._init_params for e in events}

    if len(self._demand_profiles) > 0:
        all_files = []
        for dp in self._demand_profiles.values():
            dp_dict = {k: v for k, v in dp.__dict__.items() if k not in ['id', 'sim', 'step_length', '_demand_headers']}
            all_files.append(dp_dict)
        object_params["demand"] = all_files if len(all_files) > 1 else all_files[0]

    if len(self._new_routes) > 0:
        object_params["routes"] = self._new_routes

    if filename == None:
        if self.scenario_name != None:
            filename = self.scenario_name
        else:
            desc = "No filename given."
            raise_error(ValueError, desc, self.curr_step)

    if len(object_params) == 0:
        desc = "Object save file '{0}' could not be saved (no objects found).".format(filename)
        raise_error(KeyError, desc, self.curr_step)

    if filename.endswith(".json"):
        w_class, w_mode = json, "w"
    elif filename.endswith(".pkl"):
        w_class, w_mode = pkl, "wb"
    else:
        filename += ".json"
        w_class, w_mode = json, "w"

    if os.path.exists(filename) and overwrite:
        if not self._suppress_warnings: raise_warning("File '{0}' already exists and will be overwritten.".format(filename), self.curr_step)
    elif os.path.exists(filename) and not overwrite:
        desc = "File '{0}' already exists and cannot be overwritten.".format(filename)
        raise_error(FileExistsError, desc, self.curr_step)

    with open(filename, w_mode) as fp:
        if w_mode == "wb": w_class.dump(object_params, fp)
        else: w_class.dump(object_params, fp, indent=json_indent)

def _load_objects(self, object_parameters: str | dict) -> None:
    """
    Load parameters of all TUD-SUMO objects created for the simulation (tracked edges/junctions, phases, controllers, events, demand, routes).
    
    Args:
        `object_parameters` (str, dict): Either dict containing object parameters or '_.json_'/'_.pkl_' filepath
    """
    
    object_parameters = validate_type(object_parameters, (str, dict), "object_parameters", self.curr_step)
    if isinstance(object_parameters, str):

        if object_parameters.endswith(".json"): r_class, r_mode = json, "r"
        elif object_parameters.endswith(".pkl"): r_class, r_mode = pkl, "rb"
        else:
            desc = "Invalid object parameter file '{0}' (must be '.json' or '.pkl).".format(object_parameters)
            raise_error(ValueError, desc, self.curr_step)

        if os.path.exists(object_parameters):
            with open(object_parameters, r_mode) as fp:
                object_parameters = r_class.load(fp)
        else:
            desc = "Object parameter file '{0}' not found.".format(object_parameters)
            raise_error(FileNotFoundError, desc, self.curr_step)

    valid_params = {"edges": (str, list), "junctions": dict, "phases": dict, "controllers": dict,
                    "events": dict, "demand": (str, list), "routes": dict}
    error, desc = test_input_dict(object_parameters, valid_params, "'object parameters")
    if error != None: raise_error(error, desc, self.curr_step)
    
    if "edges" in object_parameters:
        self.add_tracked_edges(object_parameters["edges"])

    if "junctions" in object_parameters:
        self.add_tracked_junctions(object_parameters["junctions"])

    if "phases" in object_parameters:
        self.set_phases(object_parameters["phases"], overwrite=True)
    
    if "controllers" in object_parameters:
        self.add_controllers(object_parameters["controllers"])

    if "events" in object_parameters:
        self.add_events(object_parameters["events"])

    if "demand" in object_parameters:
        _ = self.load_demand_profiles(object_parameters["demand"])

    if "routes" in object_parameters:

        for r_id, route in object_parameters["routes"].items():
            self.add_route(route, r_id)

def _load_demand_profiles(self, demand_profiles: str | dict | list | tuple) -> list:
    """
    Load demand profile(s) into the simulation.

    Args:
        `demand_profiles` (str, list, tuple): Either a filename to a previously saved DemandProfile object, or list of filenames

    Returns:
        list: List of DemandProfile objects
    """

    if not isinstance(demand_profiles, (list, tuple)): demand_profiles = [demand_profiles]
    validate_list_types(demand_profiles, (str, dict), param_name='demand_profiles', curr_sim_step=self.curr_step)

    loaded = []
    for dp_f in demand_profiles:

        if isinstance(dp_f, str):
            if not os.path.exists(dp_f):
                desc = f"DemandProfile file '{dp_f}' not found."
                raise_error(FileNotFoundError, desc, self.curr_step)
            
            with open(dp_f, "rb") as fp:
                dp_dict = pkl.load(fp)

        else: dp_dict = dp_f

        demand_profile = DemandProfile(self)
        demand_profile._demand_arrs = dp_dict['_demand_arrs']
        demand_profile._vehicle_types = dp_dict['_vehicle_types']

        for vehicle_type_id, vehicle_type_data in demand_profile._vehicle_types.items():
            
            if not self.vehicle_type_exists(vehicle_type_id):
                self.add_vehicle_type(vehicle_type_id, **vehicle_type_data)

        for demand_arr in demand_profile._demand_arrs:
            routing = demand_arr[0]
            
            if isinstance(routing, str) and self.route_exists(routing) == None:
                desc = f"Route with ID '{routing}' not found."
                raise_error(KeyError, desc, self.curr_step)
            
            elif isinstance(routing, (list, tuple)) and len(routing) == 2:
                e_ids = {e: e_id for e, e_id in zip(["Origin", "Destination"], routing)}
                for e, e_id in e_ids.items():
                    if self.geometry_exists(e_id) == None:
                        desc = f"{e} edge '{e_id}' not found."
                        raise_error(KeyError, desc, self.curr_step)

            else:
                desc = f"Invalid routing '{routing}'."
                raise_error(ValueError, desc, self.curr_step)

        demand_profile.step_length = self.step_length

        self._demand_profiles[demand_profile.id] = demand_profile

        self._manual_flow = True

        loaded.append(demand_profile)

    if len(loaded) == 0: loaded = loaded[0]
    return loaded

def _save_data(self, filename: str | None = None, overwrite: bool = True, json_indent: int | None = 4) -> None:
    """
    Save all vehicle, detector and junction data in a JSON or pickle file.
    
    Args:
        `filename` (str, optional): Output filepath (defaults to '_./{scenario_name}.pkl_')
        `overwrite` (bool): Prevent previous outputs being overwritten
        `json_indent` (int, optional): Indent used when saving JSON files
    """

    filename, w_class, w_mode, warning, error = _get_save_filename(self, filename, overwrite)

    if warning != None: raise_warning(warning, self.curr_step)
    elif error != None:
        desc = f"File '{filename}' already exists and cannot be overwritten."
        raise_error(error, desc, self.curr_step)

    if self._scheduler != None: self._all_data["data"]["events"] = self._scheduler.__dict__()
    with open(filename, w_mode) as fp:
        if w_mode == "wb": w_class.dump(self._all_data, fp)
        else: w_class.dump(self._all_data, fp, indent=json_indent)
        
def _save_fc_data(self, filename: str | None = None, overwrite: bool = True, json_indent: int | None = 4) -> None:
    """
    Save all floating car data in a JSON or pickle file.

    Args:
        `filename` (str, optional): Output filepath (defaults to '_./{scenario_name}_fc_data.pkl_')
        `overwrite` (bool): Prevent previous outputs being overwritten
        `json_indent` (int, optional): Indent used when saving JSON files
    """

    if not self._get_fc_data:
        desc = "No floating car data collected."
        raise_error(KeyError, desc, self.curr_step)

    filename, w_class, w_mode, warning, error = _get_save_filename(self, filename, overwrite, "_fc_data")

    if warning != None: raise_warning(warning, self.curr_step)
    elif error != None:
        desc = f"File '{filename}' already exists and cannot be overwritten."
        raise_error(error, desc, self.curr_step)

    static_keys = ["type", "departure", "destination", "origin"]
    dynamic_keys = ["longitude", "latitude", "altitude", "heading", "speed", "lane_id"]

    new_data, veh_info = {}, {}
    for key, value in self._all_data.items():
        if key != "data": new_data[key] = value
        else: 

            fc_data = []

            for step in value["fc_data"]:
                step_data = {}
                for veh_id, veh_data in step.items():
                    if veh_id not in veh_info: veh_info[veh_id] = {sk: veh_data[sk] for sk in static_keys}
                    step_data[veh_id] = {dk: veh_data[dk] for dk in dynamic_keys}

                fc_data.append(step_data)

            new_data["veh_info"] = veh_info
            new_data["fc_data"] = fc_data


    with open(filename, w_mode) as fp:
        if w_mode == "wb": w_class.dump(new_data, fp)
        else: w_class.dump(new_data, fp, indent=json_indent)

def _get_save_filename(self, filename, overwrite, def_suffix=""):

    if self._all_data == None:
        desc = "No data to save as a simulation has not been run."
        raise_error(SimulationError, desc, self.curr_step)

    if filename == None:
        if self.scenario_name != None:
            filename = self.scenario_name
        else:
            desc = "No filename given."
            raise_error(ValueError, desc, self.curr_step)

    if filename.endswith(".json"):
        w_class, w_mode = json, "w"
    elif filename.endswith(".pkl"):
        w_class, w_mode = pkl, "wb"
    else:
        filename += f"{def_suffix}.pkl"
        w_class, w_mode = pkl, "wb"

    warning, error = None, None
    if os.path.exists(filename) and overwrite:
        if not self._suppress_warnings: warning = f"File '{filename}' already exists and will be overwritten."
    elif os.path.exists(filename) and not overwrite: error = FileExistsError

    return filename, w_class, w_mode, warning, error


import traci
from copy import deepcopy
from .utils import *

def _set_tl_colour(self, junction_id: str | int, colour_str: str) -> None:
    """
    Sets a junction to a colour for an indefinite amount of time. Can be used when tracking phases separately (ie. not within TUD-SUMO).
    
    Args:
        `junction_id` (str, int): Junction ID
        `colour_str` (str): Phase colour string (valid characters are ['_G_' | '_g_' | '_y_' | '_r_' | '-'])
    """
    
    if junction_id not in list(traci.trafficlight.getIDList()):
        desc = "Junction with ID '{0}' does not exist, or it does not have a traffic light.".format(junction_id)
        raise_error(KeyError, desc, self.curr_step)
    else:
        if junction_id in self.tracked_junctions.keys():
            m_len = self.tracked_junctions[junction_id]._m_len
        else:
            state_str = traci.trafficlight.getRedYellowGreenState(junction_id)
            m_len = len(state_str)

    colour_str = validate_type(colour_str, str, "colour_str", self.curr_step)
    
    if len(colour_str) == 1:
        junc_phases = {junction_id: {"phases": [colour_str*m_len], "times": [math.inf]}}
    elif len(colour_str) == m_len:
        junc_phases = {junction_id: {"phases": [colour_str], "times": [math.inf]}}
    else:
        desc = "Invalid colour_str (must be char or len(str) == junction movements length)."
        raise_error(ValueError, desc, self.curr_step)
    
    _set_phases(self, junc_phases, overwrite=False)
    
def _set_phases(self, junction_phases: dict, start_phase: int = 0, overwrite: bool = True) -> None:
    """
    Sets the phases for the simulation, starting at the next simulation step.
    
    Args:
        `junction_phases` (dict): Dictionary containing junction phases and times
        `start_phase` (int): Phase number to start at, defaults to 0
        `overwrite` (bool): If `True`, the `junc_phases` dict is overwitten with `junction_phases`. If `False`, only specific junctions are overwritten.
    """

    # If overwriting, the junc phases dictionary is replaced with
    # the new version. Otherwise, only specific junctions are overwritten.
    if overwrite or self._junc_phases == None:
        self._junc_phases = junction_phases
    else:
        for junc_id, new_phases in junction_phases.items():
            self._junc_phases[junc_id] = deepcopy(new_phases)

    for junc_id in junction_phases.keys():

        if junc_id not in list(traci.trafficlight.getIDList()):
            desc = "Junction with ID '{0}' does not exist, or it does not have a traffic light.".format(junc_id)
            raise_error(KeyError, desc, self.curr_step)

        junc_phase = self._junc_phases[junc_id]

        valid_params = {"times": list, "phases": list, "curr_phase": int}
        error, desc = test_input_dict(junc_phase, valid_params, "'{0}' phases".format(junc_id), required=["times", "phases"])
        if error != None: raise_error(error, desc, self.curr_step)

        # Check times and colours match length, are of the right type, and assert all
        # phase times are greater than the simulation step length.
        validate_list_types(junc_phase["times"], (int, float), param_name="'{0}' phase times".format(junc_id), curr_sim_step=self.curr_step)
        validate_list_types(junc_phase["phases"], str, param_name="'{0}' phase colours".format(junc_id), curr_sim_step=self.curr_step)

        if len(junc_phase["times"]) != len(junc_phase["phases"]):
            desc = "'{0}' phase colours and times do not match length ('times' {1} != 'phases' {2}).".format(junc_id, len(junc_phase["times"]), len(junc_phase["phases"]))
            raise_error(ValueError, desc, self.curr_step)

        for t in junc_phase["times"]:
            if t < self.step_length:
                desc = "Invalid phase duration (phase_dur ({0}) < resolution ({1}))\n.  - {2}".format(t, self.step_length, junc_phase)
                raise_error(ValueError, desc, self.curr_step)

        if "curr_phase" not in junc_phase.keys(): junc_phase["curr_phase"] = start_phase
        if junc_phase["curr_phase"] > len(junc_phase["phases"]): junc_phase["curr_phase"] -= len(junc_phase["phases"])

        junc_phase["curr_time"] = sum(junc_phase["times"][:junc_phase["curr_phase"]])
        junc_phase["cycle_len"] = sum(junc_phase["times"])

    _update_lights(self, list(junction_phases.keys()))

def _set_m_phases(self, junction_phases: dict, start_phase: int = 0, overwrite: bool = True) -> None:
    """
    Sets the traffic light phases for the simulation based on movements, starting at the next simulation step.
    
    Args:
        `junction_phases` (dict): Dictionary containing junction phases, times and masks for different movements
        `start_phase` (int): Phase number to start at, defaults to 0
        `overwrite` (bool): If `True`, the `junc_phases` dict is overwitten with `junction_phases`. If `False`, only specific junctions are overwritten.
    """

    new_phase_dict = {}

    for junction_id, junc_phase in junction_phases.items():

        if junction_id not in list(traci.trafficlight.getIDList()):
            desc = "Junction with ID '{0}' does not exist, or it does not have a traffic light.".format(junction_id)
            raise_error(KeyError, desc, self.curr_step)
        else:
            if junction_id in self.tracked_junctions.keys():
                m_len = self.tracked_junctions[junction_id]._m_len
            else:
                state_str = traci.trafficlight.getRedYellowGreenState(junction_id)
                m_len = len(state_str)

        valid_params = {"phases": dict, "times": dict, "masks": dict, "curr_phase": int}
        error, desc = test_input_dict(junc_phase, valid_params, "'{0}' phase dict".format(junction_id), required=["times", "phases", "masks"])
        if error != None: raise_error(error, desc, self.curr_step)

        if set(junc_phase["phases"].keys()) == set(junc_phase["times"].keys()) == set(junc_phase["masks"].keys()):
            m_keys = list(junc_phase["phases"].keys())

            valid_params = {m_key: list for m_key in m_keys}
            error, desc = test_input_dict(junc_phase["phases"], valid_params, "'{0}' phases".format(junction_id), required=True)
            if error != None: raise_error(error, desc, self.curr_step)

            cycle_length = None
            for m_key in m_keys:
                colours, times, mask = junc_phase["phases"][m_key], junc_phase["times"][m_key], junc_phase["masks"][m_key]

                validate_list_types(colours, str, param_name="junction '{0}', movement '{1}' colours".format(junction_id, m_key), curr_sim_step=self.curr_step)
                validate_list_types(times, (int, float), param_name="junction '{0}', movement '{1}' times".format(junction_id, m_key), curr_sim_step=self.curr_step)

                if not isinstance(mask, str):
                    desc = "Invalid mask in junction '{0}', movement '{1}' (mask '{2}' is '{3}', must be str).".format(junction_id, m_key, mask, type(mask).__name__)
                    raise_error(TypeError, desc, self.curr_step)
                elif len(mask) != m_len:
                    desc = "Invalid mask in junction '{0}', movement '{1}' (mask '{2}' length does not match junction '{3}').".format(junction_id, m_key, mask, m_len)
                    raise_error(ValueError, desc, self.curr_step)

                if len(colours) != len(times):
                    desc = "Invalid phases in junction '{0}', movement '{1}' (colour and time arrays are different lengths).".format(junction_id, m_key)
                    raise_error(ValueError, desc, self.curr_step)
                elif cycle_length != None and sum(times) != cycle_length:
                    desc = "Invalid phases in junction '{0}', movement '{1}' (movement cycle length '{2}' != junction cycle length '{3}').".format(junction_id, m_key, sum(times), cycle_length)
                    raise_error(ValueError, desc, self.curr_step)
                else: cycle_length = sum(times)

                valid_colours = {"G", "g", "y", "r", "-"}
                invalid_colours = list(set(colours) - valid_colours)
                if len(invalid_colours) > 0:
                    invalid_colours.sort()
                    desc = "Invalid phase colour(s) in junction '{0}', movement '{1}' (['{2}'] are invalid, must be in ['{3}']).".format(junction_id, m_key, "','".join(invalid_colours), "' | '".join(list(valid_colours)))
                    raise_error(ValueError, desc, self.curr_step)

            complete, new_junc_phases = False, {"phases": [], "times": []}

            all_phases, all_times = junc_phase["phases"], junc_phase["times"]
            curr_phases = {m_key: junc_phase["phases"][m_key][0] for m_key in m_keys}
            curr_times = {m_key: junc_phase["times"][m_key][0] for m_key in m_keys}
            masks = junc_phase["masks"]

            while not complete:

                phase_colours = _get_phase_string(curr_phases, masks)
                phase_time = min(list(curr_times.values()))

                new_junc_phases["phases"].append(phase_colours)
                new_junc_phases["times"].append(phase_time)

                for m_key in m_keys:
                    curr_times[m_key] -= phase_time

                    if curr_times[m_key] <= 0:
                        all_phases[m_key].pop(0)
                        all_times[m_key].pop(0)
                        
                        if len(all_phases[m_key]) > 0:
                            curr_phases[m_key] = all_phases[m_key][0]
                            curr_times[m_key] = all_times[m_key][0]

                complete = sum([len(phases) for phases in all_phases.values()]) == 0

            new_phase_dict[junction_id] = new_junc_phases

        else:
            desc = "Invalid phases for junction '{0}' (movement keys do not match).".format(junction_id)
            raise_error(KeyError, desc, self.curr_step)

    _set_phases(self, new_phase_dict, start_phase=start_phase, overwrite=overwrite)

def _get_phase_string(curr_phases, masks):

    m_len = len(list(masks.values())[0])
    phase_arr = ['-'] * m_len
    
    for m_key in masks.keys():
        for idx in range(m_len):
            if masks[m_key][idx] == "1":
                phase_arr[idx] = curr_phases[m_key]

    phase_str = "".join(phase_arr)
    return phase_str

def _set_tl_metering_rate(self,
                            rm_id: str,
                            metering_rate: int | float,
                            g_time: int | float = 1,
                            y_time: int | float = 1,
                            min_red: int | float = 1,
                            vehs_per_cycle: int | None = None,
                            control_interval: int | float = 60
                        ) -> dict:
    """
    Set ramp metering rate of a meter at a junction. Uses a one-car-per-green policy with a default
    1s green and yellow time, with red phase duration changed to set flow. All phase durations must
    be larger than the simulation step length.
    
    Args:
        `rm_id` (str): Ramp meter (junction) ID
        `metering_rate` (int, float): On-ramp inflow in veh/hr (from all lanes)
        `g_time` (int, float): Green phase duration (s), defaults to 1
        `y_time` (int, float): Yellow phase duration (s), defaults to 1
        `min_red` (int, float): Minimum red phase duration (s), defaults to 1
        `vehs_per_cycle` (int, optional): Number of vehicles released with each cycle, defaults to the number of lanes
        `control_interval` (int, float): Ramp meter control interval (s)
    
    Returns:
        dict: Resulting phase dictionary
    """
    
    if min([g_time, y_time, min_red]) <= self.step_length:
        desc = "Green ({0}), yellow ({1}) and minimum red ({2}) times must all be greater than sim step length ({3}).".format(g_time, y_time, min_red, self.step_length)

    if rm_id not in list(traci.trafficlight.getIDList()):
        desc = "Junction with ID '{0}' does not exist, or it does not have a traffic light.".format(rm_id)
        raise_error(KeyError, desc, self.curr_step)
    else:
        if rm_id in self.tracked_junctions.keys():
            m_len = self.tracked_junctions[rm_id]._m_len
        else:
            state_str = traci.trafficlight.getRedYellowGreenState(rm_id)
            m_len = len(state_str)

    if self.track_juncs and rm_id in self.tracked_junctions.keys():
        if len(self.tracked_junctions[rm_id]._rate_times) == 0 or self.curr_step > self.tracked_junctions[rm_id]._rate_times[-1]:
            self.tracked_junctions[rm_id]._metering_rates.append(metering_rate)
            self.tracked_junctions[rm_id]._rate_times.append(self.curr_step)
        else:
            self.tracked_junctions[rm_id]._metering_rates[-1] = metering_rate

    # Max flow for one-car-per-green
    max_flow = (3600 / (g_time + y_time + min_red))
    
    if vehs_per_cycle == None: vehs_per_cycle = m_len

    # Max flow for n-car-per-green (accounting for no. lanes)
    max_flow *= vehs_per_cycle

    # With one lane and a g_time, y_time and min_red of 1s, the meter cannot physically release
    # more than 1200 veh/hr without reducing minimum red. So, when the metering rate is above
    # this upper bound, the meter is set to green for the whole control interval.

    # This maximum flow is increased with 2 (or more) lanes as, even with 1s green time, this essentially
    # becomes a two-car-per-green policy, and so the maximum flow is doubled.
    if metering_rate > max_flow:
        phases_dict = {"phases": ["G"*m_len], "times": [control_interval]}
    elif metering_rate == 0:
        phases_dict = {"phases": ["r"*m_len], "times": [control_interval]}
    elif metering_rate < 0:
        desc = "Metering rate must be greater than 0 (set to '{0}').".format(metering_rate)
        raise_error(ValueError, desc, self.curr_step)
    else:

        # Number of vehicles to be released per control interval
        vehicles_per_ci = (metering_rate / 3600) * control_interval

        # Number of cycles needed per control interval to achieve metering rate
        n_cycles_per_ci = vehicles_per_ci / vehs_per_cycle

        # red time calculated with the number of cycles per control interval, minus g + y time
        cycle_length = control_interval / n_cycles_per_ci
        red_time = cycle_length - g_time - y_time

        phases_dict = {"phases": ["G"*m_len, "y"*m_len, "r"*m_len],
                    "times":  [g_time, y_time, red_time]}
    _set_phases(self, {rm_id: phases_dict}, overwrite=False)
    return phases_dict

def _change_phase(self, junction_id: str | int, phase_number: int) -> None:
    """
    Change to a different phase at the specified junction_id.
    
    Args:
        `junction_id` (str, int): Junction ID
        `phase_number` (int): Phase number
    """
    
    if 0 < phase_number < len(self._junc_phases[junction_id]["phases"]):
        self._junc_phases[junction_id]["curr_phase"] = phase_number
        self._junc_phases[junction_id]["curr_time"] = sum(self.junc_phase["times"][:phase_number])

        _update_lights(self, junction_id)

    else:
        desc = "Invalid phase number '{0}' (must be [0-{1}]).".format(phase_number, len(self._junc_phases[junction_id]["phases"]))
        raise_error(ValueError, desc, self.curr_step)

def _update_lights(self, junction_ids: list | str | None = None) -> None:
    """
    Update light settings for given junctions.
    
    Args:
        `junction_ids` (list, str, optional): Junction ID, or list of IDs (defaults to all)
    """

    if junction_ids is None: junction_ids = self._junc_phases.keys()
    elif isinstance(junction_ids, str): junction_ids = [junction_ids]
    junction_ids = validate_list_types(junction_ids, str, param_name="junction_ids", curr_sim_step=self.curr_step)

    for junction_id in junction_ids:
        curr_setting = traci.trafficlight.getRedYellowGreenState(junction_id)
        new_phase = self._junc_phases[junction_id]["phases"][self._junc_phases[junction_id]["curr_phase"]]
        if '-' in new_phase:
            new_phase = new_phase.split()
            new_phase = "".join([new_phase[i] if new_phase[i] != '-' else curr_setting[i] for i in range(len(new_phase))])
        traci.trafficlight.setRedYellowGreenState(junction_id, new_phase)


import traci
from .utils import *

def _add_vehicle_subscriptions(self, vehicle_ids: str | list | tuple, data_keys: str | list | tuple) -> None:
    """
    Creates a new subscription for certain variables for **specific vehicles**. Valid data keys are;
    '_speed_', '_is_stopped_', '_max_speed_', '_acceleration_', '_position_', '_altitude_', '_heading_',
    '_edge_id_', '_lane_idx_', '_route_id_', '_route_idx_', '_route_edges_'.

    Args:
        `vehicle_ids` (str, list, tuple): Vehicle ID or list of IDs
        `data_keys` (str, list, tuple): Data key or list of keys
    """

    if isinstance(data_keys, str): data_keys = [data_keys]
    data_keys = validate_list_types(data_keys, str, param_name="data_keys", curr_sim_step=self.curr_step)

    if isinstance(vehicle_ids, str): vehicle_ids = [vehicle_ids]
    vehicle_ids = validate_list_types(vehicle_ids, str, param_name="vehicle_ids", curr_sim_step=self.curr_step)
    
    for data_key in data_keys:
        error, desc = test_valid_string(data_key, list(traci_constants["vehicle"].keys()), "data key")
        if error != None: raise_error(error, desc, self.curr_step)

    for vehicle_id in vehicle_ids:
        if self.vehicle_exists(vehicle_id):
            
            # Subscriptions are added using the traci_constants dictionary in tud_sumo.utils
            subscription_vars = [traci_constants["vehicle"][data_key] for data_key in data_keys]
            if "leader_id" in subscription_vars or "leader_dist" in subscription_vars:
                if "leader_id" in subscription_vars: subscription_vars.remove("leader_id")
                if "leader_dist" in subscription_vars: subscription_vars.remove("leader_dist")
                traci.vehicle.subscribeLeader(vehicle_id, 100)
            traci.vehicle.subscribe(vehicle_id, subscription_vars)

        else:
            desc = "Unrecognised vehicle ID given ('{0}').".format(vehicle_id)
            raise_error(KeyError, desc, self.curr_step)

def _remove_vehicle_subscriptions(self, vehicle_ids: str | list | tuple) -> None:
    """
    Remove **all** active subscriptions for a vehicle or list of vehicles.
    
    Args:
        `vehicle_ids` (str, list, tuple): Vehicle ID or list of IDs
    """

    if isinstance(vehicle_ids, str): vehicle_ids = [vehicle_ids]
    vehicle_ids = validate_list_types(vehicle_ids, str, param_name="vehicle_ids", curr_sim_step=self.curr_step)

    for vehicle_id in vehicle_ids:
        if self.vehicle_exists(vehicle_id):
            traci.vehicle.unsubscribe(vehicle_id)
        else:
            desc = "Unrecognised vehicle ID given ('{0}').".format(vehicle_id)
            raise_error(KeyError, desc, self.curr_step)

def _add_detector_subscriptions(self, detector_ids: str | list | tuple, data_keys: str | list | tuple) -> None:
    """
    Creates a new subscription for certain variables for **specific detectors**. Valid data keys are;
    '_vehicle_count_', '_vehicle_ids_', '_speed_', '_halting_no_', '_occupancy_', '_last_detection_'.
    
    Args:
        `detector_id` (str, list, tuple): Detector ID or list of IDs
        `data_keys` (str, list, tuple): Data key or list of keys
    """

    if isinstance(data_keys, str): data_keys = [data_keys]
    data_keys = validate_list_types(data_keys, str, param_name="data_keys", curr_sim_step=self.curr_step)

    if isinstance(detector_ids, str): detector_ids = [detector_ids]
    detector_ids = validate_list_types(detector_ids, str, param_name="detector_ids", curr_sim_step=self.curr_step)

    for data_key in data_keys:
        error, desc = test_valid_string(data_key, list(traci_constants["detector"].keys()), "data key")
        if error != None: raise_error(error, desc, self.curr_step)
    
    for detector_id in detector_ids:
        if detector_id not in self.available_detectors.keys():
            desc = "Detector with ID '{0}' not found.".format(detector_id)
            raise_error(KeyError, desc, self.curr_step)
        else:
            detector_type = self.available_detectors[detector_id]["type"]

            match detector_type:
                case "multientryexit": d_class = traci.multientryexit
                case "inductionloop": d_class = traci.inductionloop
                case "_":
                    desc = "Only 'multientryexit' and 'inductionloop' detectors are currently supported (not '{0}').".format(detector_type)
                    raise_error(ValueError, desc, self.curr_step)

        # Subscriptions are added using the traci_constants dictionary in tud_sumo.utils
        subscription_vars = [traci_constants["detector"][data_key] for data_key in data_keys]
        d_class.subscribe(detector_id, subscription_vars)

def _remove_detector_subscriptions(self, detector_ids: str | list | tuple) -> None:
    """
    Remove **all** active subscriptions for a detector or list of detectors.
    
    Args:
        `detector_ids` (str, list, tuple): Detector ID or list of IDs
    """

    if isinstance(detector_ids, str): detector_ids = [detector_ids]
    detector_ids = validate_list_types(detector_ids, str, param_name="detector_ids", curr_sim_step=self.curr_step)

    for detector_id in detector_ids:
        if detector_id not in self.available_detectors.keys():
            desc = "Detector with ID '{0}' not found.".format(detector_id)
            raise_error(KeyError, desc, self.curr_step)
        else:
            detector_type = self.available_detectors[detector_id]["type"]

            match detector_type:
                case "multientryexit": d_class = traci.multientryexit
                case "inductionloop": d_class = traci.inductionloop
                case "_":
                    desc = "Only 'multientryexit' and 'inductionloop' detectors are currently supported (not '{0}').".format(detector_type)
                    raise_error(ValueError, desc, self.curr_step)

            d_class.unsubscribe(detector_id)

def _add_geometry_subscriptions(self, geometry_ids: str | list | tuple, data_keys: str | list | tuple) -> None:
    """
    Creates a new subscription for geometry (edge/lane) variables. Valid data keys are;
    '_vehicle_count_', '_vehicle_ids_', '_vehicle_speed_', '_halting_no_', '_occupancy_'.
    
    Args:
        `geometry_ids` (str, list, tuple): Geometry ID or list of IDs
        `data_keys` (str, list, tuple): Data key or list of keys
    """

    if isinstance(data_keys, str): data_keys = [data_keys]
    data_keys = validate_list_types(data_keys, str, param_name="data_keys", curr_sim_step=self.curr_step)

    if isinstance(geometry_ids, str): geometry_ids = [geometry_ids]
    geometry_ids = validate_list_types(geometry_ids, str, param_name="geometry_ids", curr_sim_step=self.curr_step)

    for data_key in data_keys:
        error, desc = test_valid_string(data_key, list(traci_constants["geometry"].keys()), "data key")
        if error != None: raise_error(error, desc, self.curr_step)
    
    for geometry_id in geometry_ids:
        g_name = self.geometry_exists(geometry_id)
        if g_name == "edge": g_class = traci.edge
        elif g_name == "lane": g_class = traci.lane
        else:
            desc = "Geometry ID '{0}' not found.".format(geometry_id)
            raise_error(KeyError, desc, self.curr_step)

        # Subscriptions are added using the traci_constants dictionary in tud_sumo.utils
        subscription_vars = [traci_constants["geometry"][data_key] for data_key in data_keys]
        g_class.subscribe(geometry_id, subscription_vars)

def _remove_geometry_subscriptions(self, geometry_ids: str | list | tuple) -> None:
    """
    Remove **all** active subscriptions for a geometry object or list of geometry objects.
    
    Args:
        `geometry_ids` (str, list, tuple): Geometry ID or list of IDs
    """

    if isinstance(geometry_ids, str): geometry_ids = [geometry_ids]
    geometry_ids = validate_list_types(geometry_ids, str, param_name="geometry_ids", curr_sim_step=self.curr_step)

    for geometry_id in geometry_ids:
        g_name = self.geometry_exists(geometry_id)
        if g_name == "edge": g_class = traci.edge
        elif g_name == "lane": g_class = traci.lane
        else:
            desc = "Geometry ID '{0}' not found.".format(geometry_id)
            raise_error(KeyError, desc, self.curr_step)

        g_class.unsubscribe(geometry_id)


import traci
from random import random
from .utils import *

def _add_vehicle(self,
                vehicle_id: str,
                vehicle_type: str,
                routing: str | list | tuple,
                initial_speed: str | int | float = "max",
                origin_lane: str | int = "best",
                origin_pos: str | int | float = "base") -> None:
    """
    Add a new vehicle into the simulation.
    
    Args:
        `vehicle_id` (str): ID for new vehicle, **must be unique**
        `vehicle_type` (str): Vehicle type ID for new vehicle
        `routing` (str, list, tuple): Either route ID or (2x1) list of edge IDs for origin-destination pair
        `initial_speed` (str, int, float): Initial speed at insertion, either ['_max_' | '_random_'] or number > 0
        `origin_lane` (str, int, float): Lane for insertion at origin, either ['_random_' | '_free_' | '_allowed_' | '_best_' | '_first_'] or lane index
        `origin_pos` (str, int): Longitudinal position at insertion, either ['_random_' | '_free_' | '_random_free_' | '_base_' | '_last_' | '_stop_' | '_splitFront_'] or offset
    """

    if self.vehicle_exists(vehicle_id):
        desc = "Invalid vehicle_id given '{0}' (must be unique).".format(vehicle_id)
        raise_error(ValueError, desc, self.curr_step)
    
    origin_lane = validate_type(origin_lane, (str, int), "origin_lane", self.curr_step)
    initial_speed = validate_type(initial_speed, (str, int, float), "initial_speed", self.curr_step)
    if isinstance(initial_speed, str) and initial_speed not in ["max", "random"]:
        desc = "Invalid initial_speed string given '{0}' (must be ['_max_' | '_random_']).".format(initial_speed, type(initial_speed).__name__)
        raise_error(TypeError, desc, self.curr_step)
    elif isinstance(initial_speed, (int, float)) and initial_speed < 0:
        desc = "Invalid initial_speed value given '{0}' (must be > 0).".format(initial_speed, type(initial_speed).__name__)
        raise_error(TypeError, desc, self.curr_step)

    if isinstance(initial_speed, (int, float)):
        initial_speed = convert_units(initial_speed, self._speed_unit, "m/s")

    if not self.vehicle_type_exists(vehicle_type) and vehicle_type != "default":
        desc = "Vehicle type ID '{0}' not found.".format(vehicle_type)
        raise_error(TypeError, desc, self.curr_step)

    routing = validate_type(routing, (str, list, tuple), "routing", self.curr_step)
    if isinstance(routing, str):
        route_id = routing
        routing = self.route_exists(route_id)
        if routing != None:
            if vehicle_type != "default": traci.vehicle.add(vehicle_id, route_id, vehicle_type, departLane=origin_lane, departSpeed=initial_speed, departPos=origin_pos)
            else: traci.vehicle.add(vehicle_id, route_id, departLane=origin_lane, departSpeed=initial_speed, departPos=origin_pos)
        else:
            desc = "Route ID '{0}' not found.".format(route_id)
            raise_error(KeyError, desc, self.curr_step)

    elif isinstance(routing, (list, tuple)):
        routing = validate_list_types(routing, str, param_name="routing", curr_sim_step=self.curr_step)
        if len(routing) == 2:
            if not self.is_valid_path(routing):
                desc = "No route between edges '{0}' and '{1}'.".format(routing[0], routing[1])
                raise_error(ValueError, desc, self.curr_step)

            for geometry_id in routing:
                g_class = self.geometry_exists(geometry_id)
                if g_class == "lane":
                    desc = "Invalid geometry type (Edge ID required, '{0}' is a lane).".format(geometry_id)
                    raise_error(TypeError, desc, self.curr_step)
                
                if isinstance(origin_lane, int):
                    n_lanes = self.get_geometry_vals(geometry_id, "n_lanes")
                    if origin_lane >= n_lanes or origin_lane < 0:
                        desc = "Invalid origin lane index '{0}' (must be (0 <= origin_lane < n_lanes '{1}'))".format(origin_lane, n_lanes)
                        raise_error(ValueError, desc, self.curr_step)

            route_id = "_".join(routing)

            if self.route_exists(route_id) == None:
                traci.route.add(route_id, routing)
                self._all_routes[route_id] = tuple(routing)

            if vehicle_type != "default": traci.vehicle.add(vehicle_id, route_id, vehicle_type, departLane=origin_lane, departSpeed=initial_speed, departPos=origin_pos)
            else: traci.vehicle.add(vehicle_id, route_id, departLane=origin_lane, departSpeed=initial_speed, departPos=origin_pos)
            _vehicles_in(self, vehicle_id)

        else:
            desc = "Invalid routing given '[{0}]' (must have shape (2x1)).".format(",".join(routing))
            raise_error(TypeError, desc, self.curr_step)

def _remove_vehicles(self, vehicle_ids: str | list | tuple) -> None:
    """
    Remove a vehicle or list of vehicles from the simulation.
    
    Args:
        `vehicle_ids` (str, list, tuple): List of vehicle IDs or single ID
    """
    
    if isinstance(vehicle_ids, str): vehicle_ids = [vehicle_ids]
    vehicle_ids = validate_list_types(vehicle_ids, str, param_name="vehicle_ids", curr_sim_step=self.curr_step)

    for vehicle_id in vehicle_ids:
        if self.vehicle_exists(vehicle_id):
            traci.vehicle.remove(vehicle_id, reason=traci.constants.REMOVE_VAPORIZED)
            _vehicles_out(self, vehicle_id, is_removed=True)
        else:
            desc = "Unrecognised vehicle ID given ('{0}').".format(vehicle_id)
            raise_error(KeyError, desc, self.curr_step)

def _add_vehicle_type(self,
                        vehicle_type_id: str,
                        vehicle_class: str = "passenger",
                        colour: str | list | tuple | None = None,
                        length: int | float | None = None,
                        width: int | float | None = None,
                        height: int | float | None = None,
                        max_speed: int | float | None = None,
                        speed_factor: int | float | None = None,
                        speed_dev: int | float | None = None,
                        min_gap: int | float | None = None,
                        max_acceleration: int | float | None = None,
                        max_deceleration: int | float | None = None,
                        headway: int | float | None = None,
                        imperfection: int | float | None = None,
                        max_lateral_speed: int | float | None = None,
                        emission_class: str | None = None,
                        gui_shape: str | None = None
                    ) -> None:
    """
    Adds a new vehicle type to the simulation.

    Args:
        `vehicle_type_id` (str): ID for the new vehicle type
        `vehicle_class` (str, optional): Vehicle class (defaults to passenger)
        `colour` (str, list, tuple, optional): Vehicle colour, either hex code, list of rgb/rgba values or valid SUMO colour string
        `length` (int, float, optional): Vehicle length in metres/feet
        `width` (int, float, optional): Vehicle width in metres/feet
        `height` (int, float, optional): Vehicle height in metres/feet
        `max_speed` (int, float, optional): Vehicle max speed in km/h or mph
        `speed_factor` (int, float, optional): Vehicle speed multiplier
        `speed_dev` (int, float, optional): Vehicle deviation from speed factor
        `min_gap` (int, float, optional): Minimum gap behind leader
        `max_acceleration` (int, float, optional): Maximum vehicle acceleration
        `max_deceleration` (int, float, optional): Maximum vehicle deceleration
        `headway` (int, float, optional): Desired minimum time headway in seconds
        `imperfection` (int, float, optional): Driver imperfection (0 denotes perfect driving)
        `max_lateral_speed` (int, float, optional): Maximum lateral speed when lane changing
        `emission_class` (str, optional): Vehicle emissions class ID
        `gui_shape` (str, optional): Vehicle shape in GUI (defaults to vehicle class name)
    """
    
    if self.vehicle_type_exists(vehicle_type_id):
        desc = f"Cannot create vehicle type (ID '{vehicle_type_id}' already exists)."
        raise_error(KeyError, desc, self.curr_step)

    traci.vehicletype.copy('DEFAULT_VEHTYPE', vehicle_type_id)
    self._added_vehicle_types.add(vehicle_type_id)

    if self._gui: gui_shape = vehicle_class

    self.set_vehicle_type_vals(vehicle_type_id, vehicle_class=vehicle_class, colour=colour,
                                length=length, width=width, height=height, max_speed=max_speed,
                                speed_factor=speed_factor, speed_dev=speed_dev, min_gap=min_gap,
                                max_acceleration=max_acceleration, max_deceleration=max_deceleration,
                                headway=headway, imperfection=imperfection, max_lateral_speed=max_lateral_speed,
                                emission_class=emission_class, gui_shape=gui_shape)

def _stop_vehicle(self, vehicle_id: str, duration: int | float | None = None, lane_idx: int | None = None, pos: int | float | None = None) -> None:
    """
    Stops a vehicle at a given/random position along the next edge. The vehicle will try to stay
    in the same lane when stopping if available. If not, the vehicle will stop in the outermost lane.
    An error is thrown if the vehicle has no next edge ID (i.e. it is at the end of its route).

    Args:
        `vehicle_id` (str): Vehicle ID
        `duration` (int, float, optional): Duration to stop for in seconds (if not given, the vehicle will stop indefinitely)
        `lane_idx` (int, optional): Lane index to stop in (if not given, the vehicle will try to stop in its current lane)
        `pos` (int, float, optional): Position to stop at as a percent of the edge's length (if not given, a random position is chosen)
    """

    data = self.get_vehicle_vals(vehicle_id, ["next_edge_id", "lane_idx"])
    if data["next_edge_id"] == None:
        desc = f"Vehicle '{vehicle_id}' has no next edge ID and cannot stop."
        raise_error(ValueError, desc, self.curr_step)

    edge_data = self.get_geometry_vals(data["next_edge_id"], ["n_lanes", "length"]) 

    if pos == None: pos = random()
    elif not isinstance(pos, (int, float)):
        desc = f"Invalid pos value '{pos}' (must be [int | float], not '{type(pos).__name__}')."
        raise_error(TypeError, desc, self.curr_step)
    elif pos < 0 or pos > 1:
        desc = f"Invalid pos value '{pos}' (must be 0 <= pos < 1)."
        raise_error(ValueError, desc, self.curr_step)

    if lane_idx == None:
        lane_idx = min(data["lane_idx"], edge_data["n_lanes"] - 1)
    elif not isinstance(lane_idx, int):
        desc = f"Invalid lane_idx value '{lane_idx}' (must be int, not '{type(lane_idx).__name__}')."
        raise_error(TypeError, desc, self.curr_step)
    elif lane_idx < 0 or lane_idx >= edge_data["n_lanes"]:
        desc = f"Invalid lane_idx value '{lane_idx}' (must be 0 <= lane_idx < {edge_data['n_lanes']})."
        raise_error(ValueError, desc, self.curr_step)

    pos = pos * convert_units(edge_data["length"], self._l_dist_unit, "metres")
    if duration == None: traci.vehicle.setStop(vehicle_id, data["next_edge_id"], pos=pos, laneIndex=lane_idx)
    else: traci.vehicle.setStop(vehicle_id, data["next_edge_id"], pos=pos, laneIndex=lane_idx, duration=float(duration))
    
    self._stopped_vehicles.add(vehicle_id)

def _resume_vehicle(self, vehicle_id: str) -> None:
    """
    Resumes a previously stopped vehicle.
    
    Args:
        `vehicle_id` (str): Vehicle ID
    """

    if not self.vehicle_exists(vehicle_id):
        desc = f"Vehicle with ID '{vehicle_id}' not found."
        raise_error(KeyError, desc, self.curr_step)

    if vehicle_id not in self._stopped_vehicles and not self._suppress_warnings:
        raise_warning(f"Vehicle '{vehicle_id}' is not stopped.")

    else:
        if self.get_vehicle_vals(vehicle_id, "is_stopped"):
            traci.vehicle.resume(vehicle_id)
            self._stopped_vehicles.remove(vehicle_id)
        
        elif not self._suppress_warnings:
            raise_warning(f"Vehicle '{vehicle_id}' has not stopped yet (cannot resume).")


# --- VEHICLE FUNCTIONS ---

def _add_v_func(self, functions, parameters: dict, func_arr: list, valid_sim_params: list) -> None:
    """
    Add a vehicle in/out function.
    
    Args:
        `functions` (function, list):  Function or list of functions
        `parameters` (dict): Dictionary containing values for extra custom parameters
        `func_arr` (list):   Either list of _v_in_funcs or _v_out_funcs
        `valid_sim_params` (list): List of valid simulation parameters
    """

    if not isinstance(functions, list):
        if parameters != None: parameters = {functions.__name__: parameters}
        functions = [functions]

    for function in functions:
        func_params_arr = list(function.__code__.co_varnames)
        self._v_func_params[function.__name__] = {}
        func_arr.append(function)

        if parameters != None and function.__name__ in parameters:
            if len(set(parameters[function.__name__].keys()) - set(func_params_arr)) != 0:
                desc = "Unknown function parameters (['{0}'] are not parameter(s) of '{1}()').".format("','".join(list(set(parameters[function.__name__].keys()) - set(func_params_arr))), function.__name__)
                raise_error(KeyError, desc, self.curr_step)

        # Parameters for each function are either:
        #  - Valid simulation parameter (such as vehicle_id), which is set with each vehicle
        #  - In the extra parameters dictionary, so the value is set here
        #  - Not given, where it is assumed this has a default value already defined (error thrown if not the case)
        for func_param in func_params_arr:
            if func_param in valid_sim_params:
                self._v_func_params[function.__name__][func_param] = None
            elif parameters != None and function.__name__ in parameters and func_param in parameters[function.__name__]:
                self._v_func_params[function.__name__][func_param] = parameters[function.__name__][func_param]

def _remove_v_func(self, functions, v_func_type: str) -> None:
    """
    Remove a vehicle in/out function.
    
    Args:
        `functions` (function, list): Function (or function name) or list of functions
        `v_func_type` (str): Either 'in' or 'out'
    """
    
    if not isinstance(functions, list): functions = [functions]
    rm_func_names = [func.__name__ if not isinstance(func, str) else func for func in functions]

    for func_name in rm_func_names:
        if func_name in self._v_func_params:
            del self._v_func_params[func_name]
        else:
            desc = "Function '{0}()' not found.".format(func_name)
            raise_error(KeyError, desc, self.curr_step)

    func_arr = self._v_in_funcs if v_func_type.upper() == "IN" else self._v_out_funcs
    new_v_funcs = []
    for func in func_arr:
        if func.__name__ not in rm_func_names:
            new_v_funcs.append(func)

    if v_func_type.upper() == "IN": self._v_in_funcs = new_v_funcs
    elif v_func_type.upper() == "OUT": self._v_out_funcs = new_v_funcs

def _vehicles_in(self, vehicle_ids: str | list | tuple, is_added: bool = False) -> None:
    """
    Updates simulation with each new vehicle.

    Args:
        `vehicle_ids` (str, list, tuple): Vehicle ID or list of IDs
        `is_added` (bool): Denotes whether the vehicle(s) has been manually added
    """

    if isinstance(vehicle_ids, str): vehicle_ids = [vehicle_ids]
    elif isinstance(vehicle_ids, set): vehicle_ids = list(vehicle_ids)
    vehicle_ids = validate_list_types(vehicle_ids, str, param_name="vehicle_ids", curr_sim_step=self.curr_step)

    for vehicle_id in vehicle_ids:

        if vehicle_id not in self._all_curr_vehicle_ids: self._all_curr_vehicle_ids.add(vehicle_id)
        if vehicle_id not in self._all_loaded_vehicle_ids: self._all_loaded_vehicle_ids.add(vehicle_id)
        if vehicle_id not in self._all_added_vehicles and is_added: self._all_added_vehicles.add(vehicle_id)

        # Create a new incomplete trip in the trip data
        veh_data = self.get_vehicle_vals(vehicle_id, ("type", "route_id", "route_edges"))
        veh_type, route_id, origin, destination = veh_data["type"], veh_data["route_id"], veh_data["route_edges"][0], veh_data["route_edges"][-1]
        self._trips["incomplete"][vehicle_id] = {"route_id": route_id,
                                                    "vehicle_type": veh_type,
                                                    "departure": self.curr_step,
                                                    "origin": origin,
                                                    "destination": destination}
        
        # Call vehicle in functions
        for func in self._v_in_funcs:
            param_dict, trip_data = {}, self._trips["incomplete"][vehicle_id]
            for param, val in self._v_func_params[func.__name__].items():
                if param in trip_data: param_dict[param] = trip_data[param]
                elif param == "simulation": param_dict[param] = self
                elif param == "vehicle_id": param_dict[param] = vehicle_id
                elif param == "curr_step": param_dict[param] = self.curr_step
                else: param_dict[param] = val
            func(**param_dict)
        
def _vehicles_out(self, vehicle_ids: str | list | tuple, is_removed: bool = False) -> None:
    """
    Updates simulation with each leaving vehicle.

    Args:
        `vehicle_ids` (str, list, tuple): Vehicle IDs or list of IDs
        `is_removed` (bool): Denotes whether the vehicle(s) has been manually removed
    """

    if isinstance(vehicle_ids, str): vehicle_ids = [vehicle_ids]
    elif isinstance(vehicle_ids, set): vehicle_ids = list(vehicle_ids)
    vehicle_ids = validate_list_types(vehicle_ids, str, param_name="vehicle_ids", curr_sim_step=self.curr_step)

    for vehicle_id in vehicle_ids:

        if self.vehicle_loaded(vehicle_id) and self.vehicle_to_depart(vehicle_id): continue
        
        if is_removed or vehicle_id in self._all_removed_vehicles:
            if vehicle_id in self._trips["incomplete"].keys():
                self._trips["incomplete"][vehicle_id]["removal"] = self.curr_step
            if vehicle_id not in self._all_removed_vehicles:
                self._all_removed_vehicles.add(vehicle_id)
            if vehicle_id in self._all_curr_vehicle_ids:
                self._all_curr_vehicle_ids.remove(vehicle_id)
            if vehicle_id in self._all_loaded_vehicle_ids:
                self._all_loaded_vehicle_ids.remove(vehicle_id)
            continue

        if vehicle_id in self._trips["incomplete"].keys():
            trip_data = self._trips["incomplete"][vehicle_id]
            veh_type, route_id, departure, origin, destination = trip_data["vehicle_type"], trip_data["route_id"], trip_data["departure"], trip_data["origin"], trip_data["destination"]
            del self._trips["incomplete"][vehicle_id]
            self._trips["completed"][vehicle_id] = {"route_id": route_id, "vehicle_type": veh_type, "departure": departure, "arrival": self.curr_step, "origin": origin, "destination": destination}
        
        else:
            if vehicle_id in self._known_vehicles.keys():
                trip_data = {}
                if "departure" in self._known_vehicles[vehicle_id].keys():
                    trip_data["departure"] = self._known_vehicles[vehicle_id]["departure"]
                if "arrival" in self._known_vehicles[vehicle_id].keys():
                    trip_data["arrival"] = self._known_vehicles[vehicle_id]["arrival"]
                if "origin" in self._known_vehicles[vehicle_id].keys():
                    trip_data["origin"] = self._known_vehicles[vehicle_id]["origin"]
                if "destination" in self._known_vehicles[vehicle_id].keys():
                    trip_data["destination"] = self._known_vehicles[vehicle_id]["destination"]    
            else: 
                desc = "Unrecognised vehicle ID '{0}' in completed trips.".format(vehicle_id)
                raise_error(KeyError, desc, self.curr_step)

        for func in self._v_out_funcs:
            param_dict = {}
            for param, val in self._v_func_params[func.__name__].items():
                if param == "vehicle_id": param_dict[param] = vehicle_id
                elif param == "simulation": param_dict[param] = self
                elif param == "curr_step": param_dict[param] = self.curr_step
                else: param_dict[param] = val
            func(**param_dict)